#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Event handler to start or stop service for TMNAS."""

import os
import stat
import shutil

from iobase.cmdexec import (CmdExec, NULL)
import ionas.request as request
import ionas.confut as confut
import evh_config as config
import libevh.evhcmd as evhcmd
from fal_info import (FAL_SERVICECTRL_PATH, FAL_SHARES_PATH, Results, ReqTypes)
from product_conf import (CONF_TEMP_DIR, ETC_DIR, CONF_DIR, PROC_DIR,
                            MKFS_EXT4_CMD, MOUNT_CMD, UMOUNT_CMD)
from product_conf.devinfo import (VOL_INFO,)

EVENTINFO_NAME = "eventinfo"
VOLUME_NAME_PATH = "volume/name"
CATEGORY_ATTR = "category"
ACTION_ATTR = "action"
PREVIOUS_ATTR = "previous"

VOLUME_CATEGORY = "volume"
SYSTEM_CATEGORY = "system"
MOUNT_ACTION = "mount"
UNMOUNT_ACTION = "unmount"
SHUTDOWN_ACTION = "shutdown"
REBOOT_ACTION = "reboot"

SPLX_SERVICE = "splx"
SYSLOG_SERVICE = "syslog-ng"

NO_SUFFIX = ("",)
SYSLOG_SETTING_DIR = os.path.join(CONF_TEMP_DIR, "syslog-ng")
SYSLOG_CONF_DIR = os.path.join(CONF_DIR, "syslog-ng")
SYSLOG_CONF_PATH = os.path.join(ETC_DIR, "syslog-ng", "syslog-ng.conf")
SYSLOG_CONF_TMNAS = "syslog-ng.conf.part*.tmnas"

INTERNAL_MNT_POINT = VOL_INFO[config.INTERNAL_VOL]["mnt_point"]
VAR_DIR = "var"
OPT_SUBDIR = os.path.join(VAR_DIR, "opt")
TREND_MICRO_DIR = "TrendMicro"
LOG_SUBDIR = os.path.join(OPT_SUBDIR, TREND_MICRO_DIR, "log")
LOOPBACK_SUBDIR = os.path.join(VAR_DIR, "loopback")
NORMAL_DIR_PERM = (stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP |
                        stat.S_IROTH | stat.S_IXOTH)
TREND_MICRO_DIR_PERM = stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO

PROC_MOUNTS = os.path.join(PROC_DIR, "mounts")

LOST_FOUND_DIR = "lost+found"


SERVICECTRL_REQ_XML_TMPL = """<servicectrl>
    <servicelist>
        <service><name>%s</name></service>
    </servicelist>
</servicectrl>
"""

SHARES_LOAD_REQ_XML = "<shares/>"
SHARE_PATH = "sharelist/share"
NAME_NAME = "name"
PATH_NAME = "path"

TMADMIN_NAME = "tmadmin"
WRITE_DATA_SIZE = 128 * 1024    # 128KB
TMNAS_LOOPBACK_NAME = "tmnas.bin"
TMNAS_LOOPBACK_SIZE = 1 * 1024 * 1024 * 1024    # 1GB
TMADMIN_LOOPBACK_NAME = "tmadmin.bin"
TMADMIN_LOOPBACK_SIZE = 4 * 1024 * 1024 * 1024  # 4GB

LOOP_MOUNT_ARGS_BASE = [MOUNT_CMD, "-o", "loop"]


class EvhTmNasService(evhcmd.EvhCmd):
    """Event handler to start or stop services for TMNAS."""

    def __get_share_path(self, name):
        """Return "name" share path."""

        res_entry = self.__send_request(FAL_SHARES_PATH, ReqTypes.LOAD,
                                        SHARES_LOAD_REQ_XML)
        for share in res_entry.get().get_node_list(SHARE_PATH):
            if share.get_value(NAME_NAME) != name:
                continue
            return share.get_value(PATH_NAME)

        return ""

    def __mkdir(self, path, mode):
        """Make "path" directory including the parent directories."""

        try:
            if not os.path.exists(path):
                os.makedirs(path, mode)
        except:
            # Ignore error
            pass

    def __try_create_loopback_file(self, loopback_file, size):
        """Try creating loopback file if it does not exist."""

        if os.path.exists(loopback_file):
            # Do nothing
            return

        data = "\0" * WRITE_DATA_SIZE

        with open(loopback_file, "w") as loopback:
            wrote_total = 0
            while wrote_total < size:
                if size - wrote_total > WRITE_DATA_SIZE:
                    wrote_size = WRITE_DATA_SIZE
                else:
                    wrote_size = size - wrote_total
                loopback.write(data[0:wrote_size])
                wrote_total += wrote_size

        # Format with Ext4 filesystem.
        CmdExec(MKFS_EXT4_CMD, loopback_file).execute(stdout=NULL, stderr=NULL)

    def __prepair_dirs(self, mnt_point):
        """Prepair directories for TMANS."""

        opt_dir = os.path.join(mnt_point, OPT_SUBDIR)
        log_dir = os.path.join(mnt_point, LOG_SUBDIR)
        trend_dir = os.path.join(opt_dir, TREND_MICRO_DIR)
        loopback_dir = os.path.join(mnt_point, LOOPBACK_SUBDIR)

        old_umask = os.umask(0)

        self.__mkdir(opt_dir, NORMAL_DIR_PERM)
        self.__mkdir(trend_dir, TREND_MICRO_DIR_PERM)
        self.__mkdir(log_dir, NORMAL_DIR_PERM)
        self.__mkdir(loopback_dir, NORMAL_DIR_PERM)

        self.__try_create_loopback_file(
                            os.path.join(loopback_dir, TMNAS_LOOPBACK_NAME), 
                            TMNAS_LOOPBACK_SIZE)
        self.__try_create_loopback_file(
                            os.path.join(loopback_dir, TMADMIN_LOOPBACK_NAME),
                            TMADMIN_LOOPBACK_SIZE)

        os.umask(old_umask)

    def __mount_loopback(self, loopback_file, mnt_point):
        """Mount a loopback file."""

        if self.__is_mounted(mnt_point):
            return

        args = LOOP_MOUNT_ARGS_BASE + [loopback_file, mnt_point]
        CmdExec(args).execute(stdout=NULL, stderr=NULL)

    def __unmount_loopback(self, mnt_point):
        """Unmount a "mnt_point"."""

        CmdExec(UMOUNT_CMD, mnt_point).execute(stdout=NULL, stderr=NULL)

    def __is_mounted(self, mnt_point):
        """Check if some device is mounted to "mnt_point"."""

        with open(PROC_MOUNTS) as mounts:
            for line in mounts:
                if mnt_point == line.split(" ")[1]:
                    return True
        return False

    def __mount_loopbacks_for_tmnas(self, mnt_point):
        """Mount a loopback files for TMNAS."""

        opt_dir = os.path.join(mnt_point, OPT_SUBDIR)
        loopback_dir = os.path.join(mnt_point, LOOPBACK_SUBDIR)
        self.__mount_loopback(os.path.join(loopback_dir, TMNAS_LOOPBACK_NAME), 
                            os.path.join(opt_dir, TREND_MICRO_DIR)) 

        tmadmin_share_path = self.__get_share_path(TMADMIN_NAME)
        self.__mount_loopback(
                            os.path.join(loopback_dir, TMADMIN_LOOPBACK_NAME), 
                            tmadmin_share_path)

        if os.path.exists(os.path.join(tmadmin_share_path, LOST_FOUND_DIR)):
            shutil.rmtree(os.path.join(tmadmin_share_path, LOST_FOUND_DIR))

    def __unmount_loopbacks_for_tmnas(self, mnt_point):
        """Unmount a loopback files for TMNAS."""

        opt_dir = os.path.join(mnt_point, OPT_SUBDIR)
        self.__unmount_loopback(os.path.join(opt_dir, TREND_MICRO_DIR)) 
        self.__unmount_loopback(
                            os.path.join(self.__get_share_path(TMADMIN_NAME)))

    def __send_request(self, target, type, body):
        """Send request entry to "target".""" 

        entry = request.Entry(type, body)
        return self.send_entry(entry, target)

    def __service_control(self, name, type):
        """Control service."""

        self.__send_request(FAL_SERVICECTRL_PATH, type,
                        SERVICECTRL_REQ_XML_TMPL % (name,))

    def __update_syslog_conf(self, is_tmnas_enabled):
        """Update a configuration file for syslog."""

        if is_tmnas_enabled:
            try:
                confut.enable_setting((SYSLOG_CONF_TMNAS,), NO_SUFFIX,
                                    SYSLOG_CONF_DIR, SYSLOG_SETTING_DIR)
            except:
                self.exception("Fail enabling %s" % (SYSLOG_CONF_TMNAS,))
        else:
            try:
                confut.disable_setting((SYSLOG_CONF_TMNAS,),
                                    SYSLOG_SETTING_DIR)
            except:
                self.exception("Fail disabling %s" % (SYSLOG_CONF_TMNAS,))

    def __handle_volume_event(self, eventinfo, entry):
        """Handle volume event."""

        action = eventinfo.get_attr_value(ACTION_ATTR)
        is_previous = eventinfo.is_set_to_true(PREVIOUS_ATTR)

        if (eventinfo.get_node(VOLUME_NAME_PATH).get_text() !=
                config.INTERNAL_VOL):

            # Non target events
            entry.set_result(Results.OK)
            return entry

        if action == MOUNT_ACTION:
            if not is_previous:
                self.__prepair_dirs(INTERNAL_MNT_POINT)
                self.__mount_loopbacks_for_tmnas(INTERNAL_MNT_POINT)
                self.__service_control(SYSLOG_SERVICE, ReqTypes.STOP)
                self.__update_syslog_conf(True)
                self.__service_control(SYSLOG_SERVICE, ReqTypes.START)
                self.__service_control(SPLX_SERVICE, ReqTypes.START)

        elif action == UNMOUNT_ACTION:
            if is_previous:
                self.__service_control(SPLX_SERVICE, ReqTypes.STOP)
                self.__service_control(SYSLOG_SERVICE, ReqTypes.STOP)
                self.__unmount_loopbacks_for_tmnas(INTERNAL_MNT_POINT)
                self.__update_syslog_conf(False)
                self.__service_control(SYSLOG_SERVICE, ReqTypes.START)

        entry.set_result(Results.OK)
        return entry

    def __handle_system_event(self, eventinfo, entry):
        """Handle system event."""

        action = eventinfo.get_attr_value(ACTION_ATTR)
        is_previous = eventinfo.is_set_to_true(PREVIOUS_ATTR)

        if action in (SHUTDOWN_ACTION, REBOOT_ACTION) and is_previous:
            self.__service_control(SPLX_SERVICE, ReqTypes.STOP)
            self.__service_control(SYSLOG_SERVICE, ReqTypes.STOP)
            self.__update_syslog_conf(False)
            self.__service_control(SYSLOG_SERVICE, ReqTypes.START)

        entry.set_result(Results.OK)
        return entry

    def __event_handler(self, entry):
        """Handle "event" request."""

        try:
            eventinfo = entry.get().get_node(EVENTINFO_NAME)
            if eventinfo.get_attr_value(CATEGORY_ATTR) == VOLUME_CATEGORY:
                return self.__handle_volume_event(eventinfo, entry)

            if eventinfo.get_attr_value(CATEGORY_ATTR) == SYSTEM_CATEGORY:
                return self.__handle_system_event(eventinfo, entry)

        except:
            # Non target events
            pass

        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        evhcmd.EvhCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.EVENT, self.__event_handler)


#
# Main
#

if __name__ == "__main__":
    EvhTmNasService(option=evhcmd.EvhCmdOption()).handle_request()
